'''
VoTT to SA conversion method
'''
import logging
import json
import threading

from ..sa_json_helper import _create_vector_instance, _create_sa_json

from ....common import write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


def vott_to_sa(file_list, task, output_dir):
    classes = []
    if task == 'object_detection':
        instance_types = ['RECTANGLE']
    elif task == 'instance_segmentation':
        instance_types = ['POLYGON']
    elif task == 'vector_annotation':
        instance_types = ['RECTANGLE', 'POLYGON']

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(file_list), images_converted, images_not_converted, finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for json_file in file_list:
        json_data = json.load(open(json_file))
        file_name = '%s___objects.json' % json_data['asset']['name']
        sa_metadata = {
            'name': json_data['asset']['name'],
            'width': json_data['asset']['size']['width'],
            'height': json_data['asset']['size']['height']
        }

        instances = json_data['regions']
        sa_instances = []
        for instance in instances:
            for tag in instance['tags']:
                classes.append(tag)

            if instance['type'] in instance_types:
                if instance['type'] == 'RECTANGLE':
                    instance_type = 'bbox'
                    points = (
                        instance['boundingBox']['left'],
                        instance['boundingBox']['top'],
                        instance['boundingBox']['left'] +
                        instance['boundingBox']['width'],
                        instance['boundingBox']['top'] +
                        instance['boundingBox']['height']
                    )
                elif instance['type'] == 'POLYGON':
                    instance_type = 'polygon'
                    points = []
                    for point in instance['points']:
                        points.append(point['x'])
                        points.append(point['y'])

                sa_obj = _create_vector_instance(
                    instance_type, points, {}, [], instance['tags'][0]
                )
                sa_instances.append(sa_obj.copy())
        images_converted.append(json_data['asset']['name'])
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()
    return set(classes)



from pathlib import Path
import superannotate as sa
from ..common import upload_project


def test_voc_vector_instance(tmpdir):
    project_name = "voc2sa_vector_instance"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "VOC" / "input" / "fromPascalVOCToSuperAnnotate" / "VOC2012"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VOC", "", "Vector", "instance_segmentation"
    )

    description = 'voc vector instance segmentation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_voc_vector_object(tmpdir):
    project_name = "voc2sa_vector_object"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "VOC" / "input" / "fromPascalVOCToSuperAnnotate" / "VOC2012"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VOC", "", "Vector", "object_detection"
    )

    description = 'voc vector object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_voc_pixel(tmpdir):
    project_name = "voc2sa_pixel_instance"
    input_dir = Path(
        "tests"
    ) / "converter_test" / "VOC" / "input" / "fromPascalVOCToSuperAnnotate" / "VOC2012"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VOC", "", "Pixel", "instance_segmentation"
    )

    description = 'voc pixel instance segmentation'
    ptype = 'Pixel'
    upload_project(out_dir, project_name, description, ptype)



'''
Supervisely to SA conversion method
'''
import logging
import threading
from pathlib import Path
import json

from .supervisely_helper import _base64_to_polygon, _create_attribute_list

from ..sa_json_helper import _create_vector_instance, _create_sa_json

from ....common import write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


def supervisely_to_sa(json_files, class_id_map, task, output_dir):
    if task == 'object_detection':
        instance_types = ['rectangle']
    elif task == 'instance_segmentation':
        instance_types = ['bitmap', 'polygon']
    elif task == 'vector_annotation':
        instance_types = [
            'point', 'rectangle', 'line', 'polygon', 'cuboid', 'bitmap'
        ]

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_files), images_converted, images_not_converted,
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for json_file in json_files:
        json_data = json.load(open(json_file))
        file_name = '%s___objects.json' % Path(json_file).stem
        sa_metadata = {
            'name': Path(json_file).stem,
            'width': json_data['size']['width'],
            'height': json_data['size']['height']
        }

        sa_instances = []
        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    attributes = _create_attribute_list(
                        obj['tags'], obj['classTitle'], class_id_map
                    )

                if obj['geometryType'] in instance_types:
                    if obj['geometryType'] == 'point':
                        points = (
                            obj['points']['exterior'][0][0],
                            obj['points']['exterior'][0][1]
                        )
                        instance_type = 'point'
                    elif obj['geometryType'] == 'line':
                        instance_type = 'polyline'
                        points = [
                            item for el in obj['points']['exterior']
                            for item in el
                        ]
                    elif obj['geometryType'] == 'rectangle':
                        instance_type = 'bbox'
                        points = (
                            obj['points']['exterior'][0][0],
                            obj['points']['exterior'][0][1],
                            obj['points']['exterior'][1][0],
                            obj['points']['exterior'][1][1]
                        )
                    elif obj['geometryType'] == 'polygon':
                        instance_type = 'polygon'
                        points = [
                            item for el in obj['points']['exterior']
                            for item in el
                        ]
                    elif obj['geometryType'] == 'cuboid':
                        instance_type = 'cuboid'
                        points = (
                            obj['points'][0][0], obj['points'][0][1],
                            obj['points'][2][0], obj['points'][2][1],
                            obj['points'][4][0], obj['points'][4][1],
                            obj['points'][5][0], obj['points'][6][1]
                        )
                    elif obj['geometryType'] == 'bitmap':
                        for ppoints in _base64_to_polygon(
                            obj['bitmap']['data']
                        ):
                            points = [
                                x + obj['bitmap']['origin'][0] if i %
                                2 == 0 else x + obj['bitmap']['origin'][1]
                                for i, x in enumerate(ppoints)
                            ]
                        instance_type = 'polygon'

                    sa_obj = _create_vector_instance(
                        instance_type, points, {}, attributes, obj['classTitle']
                    )
                    sa_instances.append(sa_obj)
        images_converted.append(Path(json_file).stem)
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()


def supervisely_keypoint_detection_to_sa_vector(
    json_files, class_id_map, meta_json, output_dir
):
    classes_skeleton = {}
    for class_ in meta_json['classes']:
        if class_['shape'] != 'graph':
            continue

        classes_skeleton[class_['title']] = {'edges': [], 'nodes': {}}
        nodes = class_['geometry_config']['nodes']
        for node, value in nodes.items():
            classes_skeleton[class_['title']]['nodes'][node] = value['label']

        edges = class_['geometry_config']['edges']
        for edge in edges:
            classes_skeleton[class_['title']]['edges'].append(
                (edge['src'], edge['dst'])
            )

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_files), images_converted, images_not_converted,
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for json_file in json_files:
        file_name = '%s___objects.json' % (Path(json_file).stem)
        json_data = json.load(open(json_file))
        sa_metadata = {
            'name': Path(json_file).stem,
            'width': json_data['size']['width'],
            'height': json_data['size']['height']
        }
        sa_instances = []

        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    attributes = _create_attribute_list(
                        obj['tags'], obj['classTitle'], class_id_map
                    )

                    if obj['geometryType'] == 'graph':
                        good_nodes = []
                        nodes = obj['nodes']
                        index = 1
                        points = []
                        pointLabels = {}
                        for node, value in nodes.items():
                            good_nodes.append(node)
                            points.append(
                                {
                                    'id': index,
                                    'x': value['loc'][0],
                                    'y': value['loc'][1]
                                }
                            )
                            pointLabels[index - 1] = classes_skeleton[
                                obj['classTitle']]['nodes'][node]
                            index += 1

                        index = 1
                        connections = []
                        for edge in classes_skeleton[obj['classTitle']
                                                    ]['edges']:
                            if edge[0] not in good_nodes or edge[
                                1] not in good_nodes:
                                continue

                            connections.append(
                                {
                                    'id': index,
                                    'from': good_nodes.index(edge[0]) + 1,
                                    'to': good_nodes.index(edge[1]) + 1,
                                }
                            )
                            index += 1
                        sa_obj = _create_vector_instance(
                            'template', points, pointLabels, attributes,
                            obj['classTitle'], connections
                        )
                        sa_instances.append(sa_obj)
        images_converted.append(Path(json_file).stem)
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()



"""
"""


def _create_vector_instance(
    instance_type,
    points,
    pointLabels,
    attributes,
    class_name='',
    connections=[]
):
    sa_instance = {
        'type': instance_type,
        'pointLabels': pointLabels,
        'attributes': attributes,
        'creationType': 'Pre-annotation'
    }

    if instance_type == 'template':
        sa_instance['points'] = points
        sa_instance['connections'] = connections
    elif instance_type == 'point':
        sa_instance['x'] = points[0]
        sa_instance['y'] = points[1]
    elif instance_type == 'ellipse':
        sa_instance['cx'] = points[0]
        sa_instance['cy'] = points[1]
        sa_instance['rx'] = points[2]
        sa_instance['ry'] = points[3]
        sa_instance['angle'] = points[4]
    elif instance_type == 'bbox':
        sa_instance['points'] = {
            'x1': points[0],
            'y1': points[1],
            'x2': points[2],
            'y2': points[3]
        }
    elif instance_type in ('polygon', 'polyline'):
        sa_instance['points'] = points
    elif instance_type == 'cuboid':
        sa_instance['points'] = {
            'f1': {
                'x': points[0],
                'y': points[1]
            },
            'f2': {
                'x': points[2],
                'y': points[3],
            },
            'r1': {
                'x': points[4],
                'y': points[5]
            },
            'r2': {
                'x': points[6],
                'y': points[7]
            }
        }
    if class_name:
        sa_instance['className'] = class_name

    return sa_instance


def _create_pixel_instance(parts, attributes, class_name=''):
    sa_instance = {
        'attributes': attributes,
        'parts': parts,
    }
    if class_name:
        sa_instance['className'] = class_name

    return sa_instance


def _create_comment(points, comments):
    sa_comment = {
        'type': 'comment',
        'x': points[0],
        'y': points[1],
        'correspondence': comments
    }
    return sa_comment


def _create_sa_json(instances, metadata, tags=[], comments=[]):
    return {
        'instances': instances,
        'metadata': metadata,
        'tags': tags,
        'comments': comments
    }



import functools
import json
import logging
import os
import sys
import time
from tqdm import tqdm
from pathlib import Path
from enum import IntEnum
import numpy as np
from PIL import Image

logger = logging.getLogger("superannotate-python-sdk")

_PROJECT_TYPES = {"Vector": 1, "Pixel": 2}

_ANNOTATION_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "QualityCheck": 3,
    "Returned": 4,
    "Completed": 5,
    "Skipped": 6
}
_USER_ROLES = {"Admin": 2, "Annotator": 3, "QA": 4, "Customer": 5, "Viewer": 6}
_AVAILABLE_SEGMENTATION_MODELS = ['autonomous', 'generic']
_MODEL_TRAINING_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "Completed": 3,
    "FailedBeforeEvaluation": 4,
    "FailedAfterEvaluation": 5,
    "FailedAfterEvaluationWithSavedModel": 6
}

_PREDICTION_SEGMENTATION_STATUSES = {
    "NotStarted": 1,
    "InProgress": 2,
    "Completed": 3,
    "Failed": 4
}


def prediction_segmentation_status_from_str_to_int(status):
    return _PREDICTION_SEGMENTATION_STATUSES[status]


def prediction_segmentation_status_from_int_to_str(status):
    for idx, item in _PREDICTION_SEGMENTATION_STATUSES.items():
        if item == status:
            return idx
    raise RuntimeError("NA segmentation/prediction status")


def model_training_status_int_to_str(project_status):
    for item in _MODEL_TRAINING_STATUSES:
        if _MODEL_TRAINING_STATUSES[item] == project_status:
            return item
    raise RuntimeError("NA training status")


def model_training_status_str_to_int(project_status):
    return _MODEL_TRAINING_STATUSES[project_status]


def image_path_to_annotation_paths(image_path, project_type):
    image_path = Path(image_path)
    postfix_json = '___objects.json' if project_type == "Vector" else '___pixel.json'
    postfix_mask = '___save.png'
    if project_type == "Vector":
        return (image_path.parent / (image_path.name + postfix_json), )
    return (
        image_path.parent / (image_path.name + postfix_json),
        image_path.parent / (image_path.name + postfix_mask)
    )


def project_type_str_to_int(project_type):
    return _PROJECT_TYPES[project_type]


def project_type_int_to_str(project_type):
    """Converts metadata project_type int value to a string

    :param project_type: int in project metadata's 'type' key
    :type project_type: int

    :return: 'Vector' or 'Pixel'
    :rtype: str
    """
    for k, v in _PROJECT_TYPES.items():
        if v == project_type:
            return k
    raise RuntimeError("NA Project type")


def user_role_str_to_int(user_role):
    return _USER_ROLES[user_role]


def user_role_int_to_str(user_role):
    for k, v in _USER_ROLES.items():
        if v == user_role:
            return k
    return None


def annotation_status_str_to_int(annotation_status):
    return _ANNOTATION_STATUSES[annotation_status]


def annotation_status_int_to_str(annotation_status):
    """Converts metadata annotation_status int value to a string

    :param annotation_status: int in image metadata's 'annotation_status' key
    :type annotation_status: int

    :return: One of 'NotStarted' 'InProgress' 'QualityCheck' 'Returned' 'Completed' 'Skipped'
    :rtype: str
    """

    for k, v in _ANNOTATION_STATUSES.items():
        if v == annotation_status:
            return k
    return None


def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError(
                    '{} received both {} and {}'.format(func_name, alias, new)
                )
            logger.warning(
                '%s is deprecated; use %s in %s', alias, new, func_name
            )
            kwargs[new] = kwargs.pop(alias)


logger = logging.getLogger('httplogger')


def hex_to_rgb(hex_string):
    """Converts HEX values to RGB values
    """
    h = hex_string.lstrip('#')
    return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))


def rgb_to_hex(rgb_tuple):
    """Converts RGB values to HEX values
    """
    return '#%02x%02x%02x' % rgb_tuple


def blue_color_generator(n, hex_values=True):
    """ Blue colors generator for SuperAnnotate blue mask.
    """
    hex_colors = []
    for i in range(n + 1):
        int_color = i * 15
        bgr_color = np.array(
            [
                int_color & 255, (int_color >> 8) & 255,
                (int_color >> 16) & 255, 255
            ],
            dtype=np.uint8
        )
        hex_color = '#' + "{:02x}".format(
            bgr_color[2]
        ) + "{:02x}".format(bgr_color[1], ) + "{:02x}".format(bgr_color[0])
        if hex_values:
            hex_colors.append(hex_color)
        else:
            hex_colors.append(hex_to_rgb(hex_color))
    return hex_colors[1:]


def id2rgb(id_map):
    if isinstance(id_map, np.ndarray):
        id_map_copy = id_map.copy()
        rgb_shape = tuple(list(id_map.shape) + [3])
        rgb_map = np.zeros(rgb_shape, dtype=np.uint8)
        for i in range(3):
            rgb_map[..., i] = id_map_copy % 256
            id_map_copy //= 256
        return rgb_map
    color = []
    for _ in range(3):
        color.append(id_map % 256)
        id_map //= 256
    return color


def save_desktop_format(output_dir, classes, files_dict):
    cat_id_map = {}
    new_classes = []
    for idx, class_ in enumerate(classes):
        cat_id_map[class_['name']] = idx + 2
        class_['id'] = idx + 2
        new_classes.append(class_)
    with open(output_dir.joinpath('classes.json'), 'w') as fw:
        json.dump(new_classes, fw)

    meta = {
        "type": "meta",
        "name": "lastAction",
        "timestamp": int(round(time.time() * 1000))
    }
    new_json = {}
    files_path = []
    (output_dir / 'images' / 'thumb').mkdir()
    for file_name, json_data in files_dict.items():
        file_name = file_name.replace('___objects.json', '')
        if not (output_dir / 'images' / file_name).exists():
            continue

        for js_data in json_data:
            if 'className' in js_data:
                js_data['classId'] = cat_id_map[js_data['className']]
        json_data.append(meta)
        new_json[file_name] = json_data

        files_path.append(
            {
                'srcPath':
                    str(output_dir.resolve() / file_name),
                'name':
                    file_name,
                'imagePath':
                    str(output_dir.resolve() / file_name),
                'thumbPath':
                    str(
                        output_dir.resolve() / 'images' / 'thumb' /
                        ('thmb_' + file_name + '.jpg')
                    ),
                'valid':
                    True
            }
        )

        img = Image.open(output_dir / 'images' / file_name)
        img.thumbnail((168, 120), Image.ANTIALIAS)
        img.save(
            output_dir / 'images' / 'thumb' / ('thmb_' + file_name + '.jpg')
        )

    with open(output_dir / 'images' / 'images.sa', 'w') as fw:
        fw.write(json.dumps(files_path))

    with open(output_dir.joinpath('annotations.json'), 'w') as fw:
        json.dump(new_json, fw)

    with open(output_dir / 'config.json', 'w') as fw:
        json.dump({"pathSeparator": os.sep, "os": sys.platform}, fw)


def save_web_format(output_dir, classes, files_dict):
    for key, value in files_dict.items():
        with open(output_dir.joinpath(key), 'w') as fw:
            json.dump(value, fw, indent=2)

    with open(output_dir.joinpath('classes', 'classes.json'), 'w') as fw:
        json.dump(classes, fw)


def dump_output(output_dir, platform, classes, files_dict):
    if platform == 'Web':
        save_web_format(output_dir, classes, files_dict)
    else:
        save_desktop_format(output_dir, classes, files_dict)


def write_to_json(output_path, json_data):
    with open(output_path, 'w') as fw:
        json.dump(json_data, fw, indent=2)


MAX_IMAGE_SIZE = 100 * 1024 * 1024  # 100 MB limit
MAX_IMAGE_RESOLUTION = {
    "Vector": 100_000_000,
    "Pixel": 4_000_000
}  # Resolution limit


def process_api_response(data):
    logger.info("Maintenance operations are being performed")
    if 'metadata' not in data:
        return data

    return data['data']


def tqdm_converter(
    total_num, images_converted, images_not_converted, finish_event
):
    with tqdm(total=total_num) as pbar:
        while True:
            finished = finish_event.wait(5)
            if not finished:
                sum_all = len(images_converted) + len(images_not_converted)
                pbar.update(sum_all - pbar.n)
            else:
                pbar.update(total_num - pbar.n)
                break



'''
COCO to SA conversion method
'''

import json
import logging
import threading
from pathlib import Path

import cv2
import numpy as np

from .coco_api import _maskfrRLE, decode
from ..sa_json_helper import _create_pixel_instance, _create_sa_json
from ....common import (
    blue_color_generator, hex_to_rgb, id2rgb, write_to_json, tqdm_converter
)

logger = logging.getLogger("superannotate-python-sdk")


def annot_to_bitmask(annot):
    if isinstance(annot['counts'], list):
        bitmask = _maskfrRLE(annot)
    elif isinstance(annot['counts'], str):
        bitmask = decode(annot)

    return bitmask


def coco_panoptic_segmentation_to_sa_pixel(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    hex_colors = blue_color_generator(len(coco_json["categories"]))

    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat['name']

    img_id_to_shape = {}
    for img in coco_json['images']:
        img_id_to_shape[str(img['id'])] = {
            'height': img['height'],
            'width': img['width']
        }

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(coco_json["annotations"]), images_converted,
            images_not_converted, finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for annot in coco_json["annotations"]:
        annot_name = Path(annot["file_name"]).stem
        img_cv = cv2.imread(str(output_dir / ("%s.png" % annot_name)))
        if img_cv is None:
            images_not_converted.append("%s.jpg" % annot_name)
            logger.warning(
                "'%s' file dosen't exist!",
                output_dir / ("%s.png" % annot_name)
            )
            continue

        img = cv2.cvtColor(img_cv, cv2.COLOR_BGR2RGB)
        H, W, C = img.shape
        img = img.reshape((H * W, C))
        segments = annot["segments_info"]
        hex_colors = blue_color_generator(len(segments))

        sa_instances = []
        for i, seg in enumerate(segments):
            img[np.all(img == id2rgb(seg["id"]),
                       axis=1)] = hex_to_rgb(hex_colors[i])
            parts = [{'color': hex_colors[i]}]
            sa_obj = _create_pixel_instance(
                parts, [], cat_id_to_cat[seg["category_id"]]
            )
            sa_instances.append(sa_obj)

        img = cv2.cvtColor(img.reshape((H, W, C)), cv2.COLOR_RGB2BGR)
        cv2.imwrite(str(output_dir / ("%s.jpg___save.png" % annot_name)), img)

        images_converted.append("%s.jpg" % annot_name)
        file_name = "%s.jpg___pixel.json" % annot_name
        sa_metadata = {
            'name': annot_name,
            'width': img_id_to_shape[str(annot['image_id'])]['width'],
            'height': img_id_to_shape[str(annot['image_id'])]['height']
        }
        json_template = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, json_template)
        (output_dir / ("%s.png" % annot_name)).unlink()
    finish_event.set()
    tqdm_thread.join()


def coco_instance_segmentation_to_sa_pixel(coco_path, output_dir):
    coco_json = json.load(open(coco_path))
    cat_id_to_cat = {}
    for cat in coco_json['categories']:
        cat_id_to_cat[cat['id']] = cat

    images_dict = {}
    for img in coco_json['images']:
        images_dict[int(img['id'])] = {
            'shape': (img['height'], img['width'], 4),
            'file_name': img['file_name'],
            'annotations': []
        }

    for annot in coco_json['annotations']:
        if int(annot['image_id']) not in images_dict:
            continue
        images_dict[annot['image_id']]['annotations'].append(annot)

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(images_dict.items()), images_converted, images_not_converted,
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for id_, annotations in images_dict.items():
        file_name = '%s___pixel.json' % annotations['file_name']
        hexcolors = blue_color_generator(len(annotations['annotations']))
        mask = np.zeros(annotations['shape'])
        H, W, _ = mask.shape

        sa_instances = []
        for i, annot in enumerate(annotations['annotations']):
            hexcolor = hexcolors[i]
            color = hex_to_rgb(hexcolor)
            if isinstance(annot['segmentation'], dict):
                bitmask = annot_to_bitmask(annot['segmentation'])
                mask[bitmask == 1] = list(color)[::-1] + [255]
            else:
                for segment in annot['segmentation']:
                    bitmask = np.zeros((H, W)).astype(np.uint8)
                    pts = np.array(
                        [
                            segment[2 * i:2 * (i + 1)]
                            for i in range(len(segment) // 2)
                        ],
                        dtype=np.int32
                    )
                    cv2.fillPoly(bitmask, [pts], 1)
                    mask[bitmask == 1] = list(color)[::-1] + [255]

            parts = [{'color': hexcolor}]
            sa_obj = _create_pixel_instance(
                parts, [], cat_id_to_cat[annot['category_id']]['name']
            )
            sa_instances.append(sa_obj)

        sa_metadata = {
            'name': images_dict[id_]['file_name'],
            'width': images_dict[id_]['shape'][1],
            'height': images_dict[id_]['shape'][0]
        }
        images_converted.append(annotations['file_name'])
        json_template = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, json_template)
        cv2.imwrite(
            str(output_dir / ('%s___save.png' % annotations['file_name'])), mask
        )
    finish_event.set()
    tqdm_thread.join()



import pandas as pd


def filter_images_by_comments(
    annotations_df,
    include_unresolved_comments=True,
    include_resolved_comments=False,
    include_without_comments=False
):
    """Filter images on comment resolve status and comment existence

    :param annotations_df: pandas DataFrame of project annotations
    :type annotations_df: pandas.DataFrame
    :param include_unresolved_comments: include images with unresolved state
    :type include_unresolved_comments: bool
    :param include_resolved_comments: include images with resolved state
    :type include_resolved_comments: bool
    :param include_without_comments: include images without any comments
    :type include_without_comments: bool

    :return: filtered image names
    :rtype: list of strs

    """
    images = set()
    df = annotations_df[annotations_df["type"] == "comment"]
    if include_unresolved_comments:
        images.update(
            df[df["commentResolved"] == False]["imageName"].dropna().unique()
        )
    if include_resolved_comments:
        images.update(
            df[df["commentResolved"] == True]["imageName"].dropna().unique()
        )
    if include_without_comments:
        all_images = set(annotations_df["imageName"].dropna().unique())
        with_comments = set(df["imageName"].dropna().unique())
        images.update(all_images - with_comments)

    return list(images)


def filter_images_by_tags(annotations_df, include=None, exclude=None):
    """Filter images on tags

    :param annotations_df: pandas DataFrame of project annotations
    :type annotations_df: pandas.DataFrame
    :param include: include images with given tags
    :type include: list of strs
    :param exclude: exclude images with given tags
    :type exclude: list of strs

    :return: filtered image names
    :rtype: list of strs

    """

    df = annotations_df[annotations_df["type"] == "tag"]
    images = set(df["imageName"].dropna().unique())

    if include:
        include_images = set(
            df[df["tag"].isin(include)]["imageName"].dropna().unique()
        )
        images = images.intersection(include_images)

    if exclude:
        exclude_images = set(
            df[df["tag"].isin(exclude)]["imageName"].dropna().unique()
        )

        images = images.difference(exclude_images)

    return list(images)


def filter_annotation_instances(annotations_df, include=None, exclude=None):
    """Filter annotation instances from project annotations pandas DataFrame.

    include and exclude rules should be a list of rules of the following type:
    [{"className": "<className>", "type" : "<bbox, polygon,...>",
    "error": <True or False>, "attributes" : [{"name" : "<attribute_value>",
    "groupName" : "<attribute_group_name>"},...]},...]


    :param annotations_df: pandas DataFrame of project annotations
    :type annotations_df: pandas.DataFrame
    :param include: include rules
    :type include: list of dicts
    :param exclude: exclude rules
    :type exclude: list of dicts

    :return: filtered DataFrame
    :rtype: pandas.DataFrame

    """
    df = annotations_df.drop(["meta", "pointLabels"], axis=1)

    if include is not None:
        included_dfs = []
        for include_rule in include:
            df_new = df.copy()
            if "className" in include_rule:
                df_new = df_new[df_new["className"] == include_rule["className"]
                               ]
            if "attributes" in include_rule:
                for attribute in include_rule["attributes"]:
                    df_new = df_new[(
                        df_new["attributeGroupName"] == attribute["groupName"]
                    ) & (df_new["attributeName"] == attribute["name"])]
            if "type" in include_rule:
                df_new = df_new[df_new["type"] == include_rule["type"]]
            if "error" in include_rule:
                df_new = df_new[df_new["error"] == include_rule["error"]]
            included_dfs.append(df_new)

        df = pd.concat(included_dfs)

    if exclude is not None:
        for exclude_rule in exclude:
            df_new = df.copy()
            # with pd.option_context('display.max_rows', None):
            #     print("#", df_new["className"])
            if "className" in exclude_rule:
                df_new = df_new[df_new["className"] == exclude_rule["className"]
                               ]
            if "attributes" in exclude_rule:
                for attribute in exclude_rule["attributes"]:
                    df_new = df_new[
                        (df_new["attributeGroup"] == attribute["groupName"]) &
                        (df_new["attributeName"] == attribute["name"])]
            if "type" in exclude_rule:
                df_new = df_new[df_new["type"] == exclude_rule["type"]]
            if "error" in exclude_rule:
                df_new = df_new[df_new["error"] == exclude_rule["error"]]

            df = df.drop(df_new.index)

    result = annotations_df.loc[df.index]
    return result



'''
Googlecloud to SA conversion method
'''
import logging
import threading
from pathlib import Path

import cv2
import pandas as pd

from ..sa_json_helper import _create_vector_instance, _create_sa_json

from ....common import write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


def googlecloud_to_sa_vector(path, output_dir):
    df = pd.read_csv(path, header=None)
    dir_name = path.parent

    sa_jsons = {}
    classes = []
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(df), images_converted, images_not_converted, finish_event),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for _, row in df.iterrows():
        classes.append(row[2])

        file_name = row[1].split('/')[-1]
        try:
            images_converted.append(file_name)
            img = cv2.imread(str(dir_name / file_name))
            H, W, _ = img.shape
        except Exception as e:
            logger.warning("Can't open %s image.", file_name)
            images_not_converted.append(file_name)
            continue

        sa_file_name = '%s___objects.json' % Path(file_name).name

        points = (row[3] * W, row[4] * H, row[5] * W, row[8] * H)
        sa_instances = _create_vector_instance('bbox', points, {}, [], row[2])

        if sa_file_name in sa_jsons.keys():
            sa_jsons[sa_file_name]['instances'].append(sa_instances)
        else:
            sa_metadata = {
                'name': Path(file_name).name,
                'width': W,
                'height': H
            }
            sa_jsons[sa_file_name] = {
                'metadata': sa_metadata,
                'instances': [sa_instances]
            }
    finish_event.set()
    tqdm_thread.join()

    for key, value in sa_jsons.items():
        sa_json = _create_sa_json(value['instances'], value['metadata'])
        write_to_json(output_dir / key, sa_json)

    return classes



"""SuperAnnotate format annotation JSON helpers"""

import json

from .exceptions import SABaseException


def _create_empty_annotation_json():
    return {"instances": [], "metadata": {}, "comments": [], "tags": []}


def _preprocess_annotation_json(annotation_json):
    path = None
    if not isinstance(annotation_json, dict) and annotation_json is not None:
        path = annotation_json
        annotation_json = json.load(open(annotation_json))
    elif annotation_json is None:
        annotation_json = _create_empty_annotation_json()
    return (annotation_json, path)


def _postprocess_annotation_json(annotation_json, path):
    if path is not None:
        json.dump(annotation_json, open(path, "w"))
        return None
    else:
        return annotation_json


def add_annotation_comment_to_json(
    annotation_json,
    comment_text,
    comment_coords,
    comment_author,
    resolved=False
):
    """Add a comment to SuperAnnotate format annotation JSON


    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param comment_text: comment text
    :type comment_text: str
    :param comment_coords: [x, y] coords
    :type comment_coords: list
    :param comment_author: comment author email
    :type comment_author: str
    :param resolved: comment resolve status
    :type resolved: bool
    """

    if len(comment_coords) != 2:
        raise SABaseException(0, "Comment should have two values")

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type": "comment",
        "x": comment_coords[0],
        "y": comment_coords[1],
        "correspondence": [{
            "text": comment_text,
            "id": comment_author
        }],
        "resolved": resolved
    }
    annotation_json["comments"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_bbox_to_json(
    annotation_json,
    bbox,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add a bounding box annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param bbox: 4 element list of top-left x,y and bottom-right x, y coordinates
    :type bbox: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """

    if len(bbox) != 4:
        raise SABaseException(0, "Bounding boxes should have 4 float elements")

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "bbox",
        "points": {
            "x1": bbox[0],
            "y1": bbox[1],
            "x2": bbox[2],
            "y2": bbox[3]
        },
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_polygon_to_json(
    annotation_json,
    polygon,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add a polygon annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param polygon: [x1,y1,x2,y2,...] list of coordinates
    :type polygon: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """
    if len(polygon) % 2 != 0:
        raise SABaseException(
            0, "Polygons should be even length lists of floats."
        )

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "polygon",
        "points":
            polygon,
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_polyline_to_json(
    annotation_json,
    polyline,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add a polyline annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param polyline: [x1,y1,x2,y2,...] list of coordinates
    :type polyline: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """

    if len(polyline) % 2 != 0:
        raise SABaseException(
            0, "Polylines should be even length lists of floats."
        )

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "polyline",
        "points":
            polyline,
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_point_to_json(
    annotation_json,
    point,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add a point annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param point: [x,y] list of coordinates
    :type point: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """
    if len(point) != 2:
        raise SABaseException(0, "Point should be 2 element float list.")

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "point",
        "x":
            point[0],
        "y":
            point[1],
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_ellipse_to_json(
    annotation_json,
    ellipse,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add an ellipse annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param ellipse: [center_x, center_y, r_x, r_y, angle]
                    list of coordinates and rotation angle in degrees around y
                    axis
    :type ellipse: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """
    if len(ellipse) != 5:
        raise SABaseException(0, "Ellipse should be 5 element float list.")

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "ellipse",
        "cx":
            ellipse[0],
        "cy":
            ellipse[1],
        "rx":
            ellipse[2],
        "ry":
            ellipse[3],
        "angle":
            ellipse[4],
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_template_to_json(
    annotation_json,
    template_points,
    template_connections,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add a template annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param template_points: [x1,y1,x2,y2,...] list of coordinates
    :type template_points: list of floats
    :param template_connections: [from_id_1,to_id_1,from_id_2,to_id_2,...]
                                 list of indexes from -> to. Indexes are based
                                 on template_points. E.g., to have x1,y1 to connect
                                 to x2,y2 and x1,y1 to connect to x4,y4,
                                 need: [1,2,1,4,...]
    :type template_connections: list of ints
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """
    if len(template_points) % 2 != 0:
        raise SABaseException(
            0, "template_points should be even length lists of floats."
        )
    if len(template_connections) % 2 != 0:
        raise SABaseException(
            0, "template_connections should be even length lists of ints."
        )

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "template",
        "points": [],
        "connections": [],
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }
    for i in range(0, len(template_points), 2):
        annotation["points"].append(
            {
                "id": i // 2 + 1,
                "x": template_points[i],
                "y": template_points[i + 1]
            }
        )
    for i in range(0, len(template_connections), 2):
        annotation["connections"].append(
            {
                "id": i // 2 + 1,
                "from": template_connections[i],
                "to": template_connections[i + 1]
            }
        )

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_cuboid_to_json(
    annotation_json,
    cuboid,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None
):
    """Add a cuboid annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param cuboid: [x_front_tl,y_front_tl,x_front_br,y_front_br,
                    x_rear_tl,y_rear_tl,x_rear_br,y_rear_br] list of coordinates
                    of front rectangle and back rectangle, in top-left (tl) and
                    bottom-right (br) format
    :type cuboid: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of attributes
    :type error: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """
    if len(cuboid) != 8:
        raise SABaseException(0, "cuboid should be lenght 8 list of floats.")

    annotation_json, path = _preprocess_annotation_json(annotation_json)

    annotation = {
        "type":
            "cuboid",
        "points":
            {
                "f1": {
                    "x": cuboid[0],
                    "y": cuboid[1]
                },
                "f2": {
                    "x": cuboid[2],
                    "y": cuboid[3]
                },
                "r1": {
                    "x": cuboid[4],
                    "y": cuboid[5]
                },
                "r2": {
                    "x": cuboid[6],
                    "y": cuboid[7]
                },
            },
        "className":
            annotation_class_name,
        "error":
            error,
        "groupId":
            0,
        "pointLabels": {},
        "locked":
            False,
        "visible":
            True,
        "attributes":
            [] if annotation_class_attributes is None else
            annotation_class_attributes
    }

    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)



import io
import logging
import re
import time
from pathlib import Path

import boto3

from .. import common
from ..api import API
from ..exceptions import SABaseException, SAImageSizeTooLarge
from .images import (
    delete_image, get_image_annotations, get_image_bytes, get_image_metadata,
    search_images, set_image_annotation_status, upload_image_annotations
)
from .project_api import get_project_metadata_bare
from .projects import (
    __create_image, get_project_default_image_quality_in_editor,
    get_image_array_to_upload, upload_image_array_to_s3
)

logger = logging.getLogger("superannotate-python-sdk")
_api = API.get_instance()


def upload_image_to_project(
    project,
    img,
    image_name=None,
    annotation_status="NotStarted",
    from_s3_bucket=None,
    image_quality_in_editor=None
):
    """Uploads image (io.BytesIO() or filepath to image) to project.
    Sets status of the uploaded image to set_status if it is not None.

    :param project: project name or metadata of the project to upload image to
    :type project: str or dict
    :param img: image to upload
    :type img: io.BytesIO() or Pathlike (str or Path)
    :param image_name: image name to set on platform. If None and img is filepath,
                       image name will be set to filename of the path
    :type image_name: str
    :param annotation_status: value to set the annotation statuses of the uploaded image NotStarted InProgress QualityCheck Returned Completed Skipped
    :type annotation_status: str
    :param from_s3_bucket: AWS S3 bucket to use. If None then folder_path is in local filesystem
    :type from_s3_bucket: str
    :param image_quality_in_editor: image quality be seen in SuperAnnotate web annotation editor.
           Can be either "compressed" or "original".  If None then the default value in project settings will be used.
    :type image_quality_in_editor: str
    """
    if not isinstance(project, dict):
        project = get_project_metadata_bare(project)
    annotation_status = common.annotation_status_str_to_int(annotation_status)
    if image_quality_in_editor is None:
        image_quality_in_editor = get_project_default_image_quality_in_editor(
            project
        )

    img_name = None
    if not isinstance(img, io.BytesIO):
        img_name = Path(img).name
        if from_s3_bucket is not None:
            from_session = boto3.Session()
            from_s3 = from_session.resource('s3')
            from_s3_object = from_s3.Object(from_s3_bucket, img)
            file_size = from_s3_object.content_length
            if file_size > common.MAX_IMAGE_SIZE:
                raise SAImageSizeTooLarge(file_size, img)
            img = io.BytesIO()
            from_s3_object.download_fileobj(img)
        else:
            file_size = Path(img).stat().st_size
            if file_size > common.MAX_IMAGE_SIZE:
                raise SAImageSizeTooLarge(file_size, img)
            with open(img, "rb") as f:
                img = io.BytesIO(f.read())
    elif img.getbuffer().nbytes > common.MAX_IMAGE_SIZE:
        raise SAImageSizeTooLarge(file_size)

    if image_name is not None:
        img_name = image_name

    if img_name is None:
        raise SABaseException(
            0, "Image name img_name should be set if img is not Pathlike"
        )

    team_id, project_id = project["team_id"], project["id"]
    params = {
        'team_id': team_id,
    }
    response = _api.send_request(
        req_type='GET',
        path=f'/project/{project_id}/sdkImageUploadToken',
        params=params
    )
    if response.ok:
        res = response.json()
        prefix = res['filePath']
    else:
        raise SABaseException(
            response.status_code, "Couldn't get upload token " + response.text
        )
    s3_session = boto3.Session(
        aws_access_key_id=res['accessKeyId'],
        aws_secret_access_key=res['secretAccessKey'],
        aws_session_token=res['sessionToken']
    )
    s3_resource = s3_session.resource('s3')
    bucket = s3_resource.Bucket(res["bucket"])
    key = prefix + f'{img_name}'
    try:
        images_array = get_image_array_to_upload(
            img, image_quality_in_editor, project["type"]
        )
        upload_image_array_to_s3(bucket, *images_array, key, project["type"])
    except Exception as e:
        raise SABaseException(0, "Couldn't upload to data server. " + e)

    __create_image([img_name], project, annotation_status, prefix)

    while True:
        try:
            get_image_metadata(project, img_name)
        except SABaseException:
            time.sleep(0.2)
        else:
            break


def copy_image(
    source_project,
    image_name,
    destination_project,
    include_annotations=False,
    copy_annotation_status=False,
    copy_pin=False
):
    """Copy image to a project. The image's project is the same as destination
    project then the name will be changed to <image_name>_(<num>).<image_ext>,
    where <num> is the next available number deducted from project image list.

    :param source_project: project name or metadata of the project of source project
    :type source_project: str or dict
    :param image_name: image name
    :type image: str
    :param destination_project: project name or metadata of the project of destination project
    :type destination_project: str or dict
    :param include_annotations: enables annotations copy
    :type include_annotations: bool
    :param copy_annotation_status: enables annotations status copy
    :type copy_annotation_status: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    """
    if not isinstance(source_project, dict):
        source_project = get_project_metadata_bare(source_project)
    if not isinstance(destination_project, dict):
        destination_project = get_project_metadata_bare(destination_project)
    img_b = get_image_bytes(source_project, image_name)
    img_metadata = get_image_metadata(source_project, image_name)
    new_name = image_name
    extension = Path(image_name).suffix
    p = re.compile(r"_\([0-9]+\)\.")
    while True:
        try:
            get_image_metadata(destination_project, new_name)
        except SABaseException:
            break
        else:
            found_copied = False
            for m in p.finditer(new_name):
                if m.start() + len(m.group()
                                  ) + len(extension) - 1 == len(new_name):
                    num = int(m.group()[2:-2])
                    found_copied = True
                    break
            if not found_copied:
                new_name = Path(new_name).stem + "_(1)" + extension
            else:
                new_name = new_name[:m.start() +
                                    2] + str(num + 1) + ")" + extension

    upload_image_to_project(destination_project, img_b, new_name)
    if include_annotations:
        annotations = get_image_annotations(source_project, image_name)
        if annotations["annotation_json"] is not None:
            if "annotation_mask" in annotations:
                upload_image_annotations(
                    destination_project, new_name,
                    annotations["annotation_json"],
                    annotations["annotation_mask"]
                )
            else:
                upload_image_annotations(
                    destination_project, new_name,
                    annotations["annotation_json"]
                )
    if copy_annotation_status:
        set_image_annotation_status(
            destination_project, new_name, img_metadata["annotation_status"]
        )
    if copy_pin:
        pin_image(destination_project, new_name, img_metadata["is_pinned"])

    logger.info(
        "Copied image %s/%s to %s/%s.", source_project["name"], image_name,
        destination_project["name"], new_name
    )


def move_image(
    source_project,
    image_name,
    destination_project,
    include_annotations=True,
    copy_annotation_status=True,
    copy_pin=True
):
    """Move image from source_project to destination_project. source_project
    and destination_project cannot be the same.

    :param source_project: project name or metadata of the project of source project
    :type source_project: str or dict
    :param image_name: image name
    :type image: str
    :param destination_project: project name or metadata of the project of destination project
    :type destination_project: str or dict
    :param include_annotations: enables annotations move
    :type include_annotations: bool
    :param copy_annotation_status: enables annotations status copy
    :type copy_annotation_status: bool
    :param copy_pin: enables image pin status copy
    :type copy_pin: bool
    """
    if not isinstance(source_project, dict):
        source_project = get_project_metadata_bare(source_project)
    if not isinstance(destination_project, dict):
        destination_project = get_project_metadata_bare(destination_project)
    if source_project == destination_project:
        raise SABaseException(
            0, "Cannot move image if source_project == destination_project."
        )
    copy_image(
        source_project, image_name, destination_project, include_annotations,
        copy_annotation_status, copy_pin
    )
    delete_image(source_project, image_name)
    logger.info("Deleted image %s/%s.", source_project["name"], image_name)


def pin_image(project, image_name, pin=True):
    """Pins (or unpins) image

    :param project: project name or metadata of the project
    :type project: str or dict
    :param image_name: image name
    :type image: str
    :param pin: sets to pin if True, else unpins image
    :type pin: bool
    """
    if not isinstance(project, dict):
        project = get_project_metadata_bare(project)
    img_metadata = get_image_metadata(project, image_name)
    team_id, project_id, image_id = project["team_id"], project[
        "id"], img_metadata["id"]
    params = {"team_id": team_id, "project_id": project_id}
    json_req = {"is_pinned": int(pin)}
    response = _api.send_request(
        req_type='PUT',
        path=f'/image/{image_id}',
        params=params,
        json_req=json_req
    )
    if not response.ok:
        raise SABaseException(
            response.status_code, "Couldn't pin image " + response.text
        )


def assign_images(project, image_names, user):
    """Assigns images to a user. The assignment role, QA or Annotator, will
    be deduced from the user's role in the project. With SDK, the user can be
    assigned to a role in the project with the share_project function.

    :param project: project name or metadata of the project
    :type project: str or dict
    :param image_names: list of image names to assign
    :type image_names: list of str
    :param user: user email
    :type user: str
    """
    logger.info("Assign %s images to user %s", len(image_names), user)
    if len(image_names) == 0:
        return
    if not isinstance(project, dict):
        project = get_project_metadata_bare(project)
    folder_id = None
    images = search_images(project, return_metadata=True)
    image_dict = {}
    for image in images:
        image_dict[image["name"]] = image["id"]
        if folder_id is None:
            folder_id = image["folder_id"]
        elif folder_id != image["folder_id"]:
            raise SABaseException(0, "Folders not implemented yet")

    image_ids = []
    for image_name in image_names:
        image_ids.append(image_dict[image_name])
    team_id, project_id = project["team_id"], project["id"]
    params = {
        "team_id": team_id,
        "project_id": project_id,
        "folder_id": folder_id
    }
    json_req = {"user_id": user, "image_ids": image_ids}
    response = _api.send_request(
        req_type='POST',
        path='/images/assign',
        params=params,
        json_req=json_req
    )
    if not response.ok:
        raise SABaseException(
            response.status_code, "Couldn't assign images " + response.text
        )
    # print(response.json())



import logging

from ..api import API
from ..exceptions import (
    SAExistingProjectNameException, SANonExistingProjectNameException
)
from .project_api import get_project_metadata_bare
from .project_metadata import get_project_metadata
from .projects import create_project_from_metadata

logger = logging.getLogger("superannotate-python-sdk")
_api = API.get_instance()


def clone_project(
    project_name,
    from_project,
    project_description=None,
    copy_annotation_classes=True,
    copy_settings=True,
    copy_workflow=True,
    copy_contributors=False
):
    """Create a new project in the team using annotation classes and settings from from_project.

    :param project_name: new project's name
    :type project_name: str
    :param from_project: the name or metadata of the project being used for duplication
    :type from_project: str or dict
    :param project_description: the new project's description. If None, from_project's
                                description will be used
    :type project_description: str
    :param copy_annotation_classes: enables copying annotation classes
    :type copy_annotation_classes: bool
    :param copy_settings: enables copying project settings
    :type copy_settings: bool
    :param copy_workflow: enables copying project workflow
    :type copy_workflow: bool
    :param copy_contributors: enables copying project contributors
    :type copy_contributors: bool

    :return: dict object metadata of the new project
    :rtype: dict
    """
    try:
        get_project_metadata_bare(project_name)
    except SANonExistingProjectNameException:
        pass
    else:
        raise SAExistingProjectNameException(
            0, "Project with name " + project_name +
            " already exists. Please use unique names for projects to use with SDK."
        )
    metadata = get_project_metadata(
        from_project, copy_annotation_classes, copy_settings, copy_workflow,
        copy_contributors
    )
    metadata["name"] = project_name
    if project_description is not None:
        metadata["description"] = project_description

    return create_project_from_metadata(metadata)



'''
'''
import logging
import threading
from pathlib import Path

from PIL import Image

from .coco_converter import CocoBaseStrategy

from ....common import id2rgb, write_to_json, tqdm_converter

logger = logging.getLogger("superannotate-python-sdk")


class CocoPanopticConverterStrategy(CocoBaseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def _sa_to_coco_single(self, id_, annotation_json, id_generator):

        image_commons = self._prepare_single_image_commons(
            id_, annotation_json['metadata']
        )
        res = self.conversion_algorithm(
            image_commons, annotation_json['instances'], id_generator
        )
        return res

    def sa_to_output_format(self):
        out_json = self._create_skeleton()
        out_json['categories'] = self._create_categories(
            self.export_root / 'classes_mapper.json'
        )

        images = []
        annotations = []
        id_generator = self._make_id_generator()
        annot_id_generator = self._make_id_generator()

        jsons = self.make_anno_json_generator()
        images_converted = []
        images_not_converted = []
        total_num = self.get_num_total_images()
        finish_event = threading.Event()
        tqdm_thread = threading.Thread(
            target=tqdm_converter,
            args=(
                total_num, images_converted, images_not_converted, finish_event
            ),
            daemon=True
        )
        logger.info('Converting to COCO JSON format')
        tqdm_thread.start()
        for json_ in jsons:
            idx = next(annot_id_generator)
            res = self._sa_to_coco_single(idx, json_, id_generator)

            panoptic_mask = json_['metadata']['panoptic_mask']

            Image.fromarray(id2rgb(res[2])).save(panoptic_mask)

            annotation = {
                'image_id': res[0]['id'],
                'file_name': Path(panoptic_mask).name,
                'segments_info': res[1]
            }
            annotations.append(annotation)

            images.append(res[0])
            images_converted.append(json_)

        out_json['annotations'] = annotations
        out_json['images'] = images
        write_to_json(
            self.output_dir / '{}.json'.format(self.dataset_name), out_json
        )
        finish_event.set()
        tqdm_thread.join()


class CocoObjectDetectionStrategy(CocoBaseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def _sa_to_coco_single(self, id_, annotation_json, id_generator):

        image_commons = self._prepare_single_image_commons(
            id_, annotation_json['metadata']
        )

        def make_annotation(
            category_id, image_id, bbox, segmentation, area, anno_id
        ):
            if self.task == 'object_detection':
                segmentation = [
                    [
                        bbox[0], bbox[1], bbox[0], bbox[1] + bbox[3],
                        bbox[0] + bbox[2], bbox[1] + bbox[3], bbox[0] + bbox[2],
                        bbox[1]
                    ]
                ]
            annotation = {
                'id': anno_id,  # making sure ids are unique
                'image_id': image_id,
                'segmentation': segmentation,
                'iscrowd': 0,
                'bbox': bbox,
                'area': area,
                'category_id': category_id
            }

            return annotation

        res = self.conversion_algorithm(
            make_annotation, image_commons, annotation_json['instances'],
            id_generator
        )
        return res

    def sa_to_output_format(self):
        out_json = self._create_skeleton()
        out_json['categories'] = self._create_categories(
            self.export_root / 'classes_mapper.json'
        )

        images = []
        annotations = []
        id_generator = self._make_id_generator()
        annot_id_generator = self._make_id_generator()
        jsons = self.make_anno_json_generator()

        images_converted = []
        images_not_converted = []
        finish_event = threading.Event()
        total_num = self.get_num_total_images()
        tqdm_thread = threading.Thread(
            target=tqdm_converter,
            args=(
                total_num, images_converted, images_not_converted, finish_event
            ),
            daemon=True
        )
        logger.info('Converting to COCO JSON format')
        tqdm_thread.start()
        for json_ in jsons:
            idx = next(annot_id_generator)
            try:
                images_converted.append(json_)
                res = self._sa_to_coco_single(idx, json_, id_generator)
            except Exception as e:
                images_not_converted.append(json_)
                raise

            images.append(res[0])
            if len(res[1]) < 1:
                self.increase_converted_count()
            for ann in res[1]:
                annotations.append(ann)
        out_json['annotations'] = annotations
        out_json['images'] = images

        write_to_json(
            self.output_dir / '{}.json'.format(self.dataset_name), out_json
        )
        finish_event.set()
        tqdm_thread.join()


class CocoKeypointDetectionStrategy(CocoBaseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def sa_to_output_format(self):
        out_json = self._create_skeleton()
        jsons = self.make_anno_json_generator()

        total_num = self.get_num_total_images()
        id_generator = self._make_id_generator()
        id_generator_anno = self._make_id_generator()
        id_generator_img = self._make_id_generator()

        res = self.conversion_algorithm(
            jsons, id_generator, id_generator_anno, id_generator_img,
            self._make_image_info, total_num
        )

        out_json['categories'] = res[0]
        out_json['annotations'] = res[1]
        out_json['images'] = res[2]
        write_to_json(
            self.output_dir / '{}.json'.format(self.dataset_name), out_json
        )
        self.set_num_converted(len(out_json['images']))



"""This document containes functions to run smart segmentation, model traing and smart prediction
develoer  : Vahagn Tumanyan
maintainer: Vahagn Tumanyan
email     : Vahagn@superannotate.com
"""
import json
import logging
import os
import time

import boto3
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from ..api import API
from ..common import (
    _AVAILABLE_SEGMENTATION_MODELS, model_training_status_int_to_str,
    process_api_response, project_type_str_to_int
)
from ..db.images import get_image_metadata, search_images
from ..exceptions import SABaseException
from ..parameter_decorators import model_metadata, project_metadata
from .defaults import DEFAULT_HYPERPARAMETERS, NON_PLOTABLE_KEYS
from .utils import log_process, make_plotly_specs, reformat_metrics_json

logger = logging.getLogger("superannotate-python-sdk")
_api = API.get_instance()


@project_metadata
@model_metadata
def run_prediction(project, images_list, model):
    """This function runs smart prediction on given list of images from a given project using the neural network of your choice
    :param project: the project in which the target images are uploaded.
    :type project: str or dict
    :param images_list: the list of image names on which smart prediction has to be run
    :type images_list: list of str
    :param model: the name of the model that should be used for running smart prediction
    :type model: str or dict
    :out res: tupe of two lists, list of images on which the prediction has succeded and failed respectively
    :rtype res: tuple
    """

    if not isinstance(project, dict):
        raise SABaseException(
            0,
            "smart prediction cannot be run on images from different projects simultaneously"
        )

    model = model.get(project['type'], None)
    if not model:
        raise SABaseException(
            0,
            f"Specified project has type {project['type']}, and does not correspond to the type of provided model"
        )
    project_id = project["id"]

    images_metadata = get_image_metadata(project, images_list)
    if isinstance(images_metadata, dict):
        images_metadata = [images_metadata]
    images_metadata.sort(key=lambda x: x['name'])

    if len(images_metadata) == 0:
        raise SABaseException(0, "No valid image names were provided")

    skipped_images_num = len(images_list) - len(images_metadata)

    if skipped_images_num > 0:
        logger.warning(
            f"{skipped_images_num} images did not exist in the provided project and were skipped."
        )

    image_name_set = set([x['name'] for x in images_metadata])
    image_id_list = [x['id'] for x in images_metadata]

    params = {
        "team_id": _api.team_id,
        "project_id": project_id,
        "image_ids": image_id_list,
        "ml_model_id": model["id"]
    }

    response = _api.send_request(
        req_type="POST", json_req=params, path="/images/prediction"
    )

    if not response.ok:
        raise SABaseException(0, "Could not start prediction")

    logger.info("Started smart prediction")

    total_image_count = len(image_name_set)
    succeded_imgs, failed_imgs = log_process(
        project, image_name_set, total_image_count, 'prediction_status',
        "Smart Prediction", logger
    )

    return succeded_imgs, failed_imgs


@project_metadata
def run_segmentation(project, images_list, model):
    """Starts smart segmentation on a list of images using the specified model

    :param project: project name of metadata of the project
    :type  project: str or dict
    :param model  : The model name or metadata of the model
    :type  model  : str or dict
    :out res: tupe of two lists, list of images on which the prediction has succeded and failed respectively
    :rtype res: tuple
    """

    if project['type'] != 'Pixel':
        logger.error(
            f"Smart segmentation is supported only for 'Pixel' projects"
        )
        raise SABaseException(
            0, "Operation not supported for given project type"
        )

    if model not in _AVAILABLE_SEGMENTATION_MODELS:
        logger.error(
            f"the model does not exist, please chose from {_AVAILABLE_SEGMENTATION_MODELS}"
        )
        raise SABaseException(0, "Model Does not exist")

    images_metadata = get_image_metadata(project, images_list)
    images_metadata.sort(key=lambda x: x["name"])

    if len(images_metadata) == 0:
        raise SABaseException(0, "No valid image names were provided")

    skipped_images_num = len(images_list) - len(images_metadata)

    if skipped_images_num > 0:
        logger.warning(
            f"{skipped_images_num} images did not exist in the provided project and were skipped."
        )

    image_name_set = set([x['name'] for x in images_metadata])
    images_id_list = [x['id'] for x in images_metadata]

    total_image_count = len(image_name_set)
    json_req = {"model_name": model, "image_ids": images_id_list}

    params = {"team_id": _api.team_id, "project_id": project["id"]}

    response = _api.send_request(
        path=f"/images/segmentation",
        req_type="POST",
        params=params,
        json_req=json_req
    )

    if not response.ok:
        logger.error("Could not start segmentation")
        raise SABaseException(0, "Could not start prediction")

    logger.info("Started smart segmentation")

    succeded_imgs, failed_imgs = log_process(
        project, image_name_set, total_image_count, 'segmentation_status',
        'Smart Segmentation', logger
    )

    return (succeded_imgs, failed_imgs)


@project_metadata
@model_metadata
def run_training(
    project,
    base_model,
    model_name,
    model_description,
    task,
    hyperparameters,
    log=False
):
    """Runs neural network training
    :param project: project or list of projects that contain the training images
    :type  project: str, dict or list of dict
    :param base_model  : base model on which the new network will be trained
    :type  base_model  : str or dict
    :param model_name   : name of the new model
    :type  model_name   : str
    :param model_description: description of the new model
    :type  model_description: str
    :param task   : The model training task
    :type  task   : str
    :param hyperparameters: hyperparameters that should be used in training
    :type  hyperparameters: dict
    :param log: If true will log training metrics in the stdout
    :type log: boolean
    :out new_model: the metadata of the newly created model
    :rtype new_model: dict
    """

    project_ids = None
    project_type = None

    if isinstance(project, dict):
        project_ids = [project["id"]]
        project_type = project["type"]
        project = [project]
    else:
        project_ids = [x["id"] for x in project]
        types = (x["type"] for x in project)
        types = set(types)
        if len(types) != 1:
            logger.error(
                "All projects have to be of the same type. Either vector or pixel"
            )
            raise SABaseException(0, "Invalid project types")
        project_type = types.pop()

    base_model = base_model.get(project_type, None)
    if not base_model:
        logger.error(
            "The base model has to be of the same type (vector or pixel) as the projects"
        )
        raise SABaseException(
            0,
            f"The type of provided projects is {project_type}, and does not correspond to the type of provided model"
        )

    for item in DEFAULT_HYPERPARAMETERS:
        if item not in hyperparameters:
            hyperparameters[item] = DEFAULT_HYPERPARAMETERS[item]
    complete_image_count = 0
    for proj in project:
        complete_image_count += proj['completedImagesCount']

    hyperparameters["name"] = model_name
    hyperparameters["description"] = model_description
    hyperparameters["task"] = task
    hyperparameters["base_model_id"] = base_model["id"]
    hyperparameters["project_ids"] = project_ids
    hyperparameters["image_count"] = complete_image_count
    hyperparameters["project_type"] = project_type_str_to_int(project_type)
    params = {
        "team_id": _api.team_id,
    }

    response = _api.send_request(
        req_type="POST",
        path="/ml_models",
        json_req=hyperparameters,
        params=params
    )
    if response.ok:
        logger.info("Started model training")
    else:
        logger.error("Could not start training")
        raise SABaseException(0, "Could not start training")
    new_model = response.json()
    new_model_id = new_model['id']
    if not log:
        return new_model

    logger.info(
        "We are firing up servers to run model training. Depending on the number of training images and the task it may take up to 15 minutes until you will start seeing metric reports"
    )
    logger.info(
        "Terminating the function will not terminate model training. If you wish to stop the training please use the stop_model_training function"
    )

    is_training_finished = False

    while not is_training_finished:
        metrics_response = _api.send_request(
            req_type='GET',
            path=f'/ml_model/{new_model_id}/getCurrentMetrics',
            params=params
        )

        metrics_data = metrics_response.json()
        if len(metrics_data) == 1:
            logger.info('Starting up servers')
            time.sleep(30)
        if 'continuous_metrics' in metrics_data:
            logger.info(metrics_data['continuous_metrics'])
        if 'per_evaluation_metrics' in metrics_data:
            for item, value in metrics_data['per_evaluation_metrics'].items():
                logger.info(value)
        if 'training_status' in metrics_data:
            status_str = model_training_status_int_to_str(
                metrics_data['training_status']
            )
            if status_str == 'Completed':
                logger.info('Model Training Successfully completed')
                is_training_finished = True
            elif status_str == 'FailedBeforeEvaluation' or status_str == 'FailedAfterEvaluation':
                logger.info('Failed to train model')
                is_training_finished = True
            elif status_str == 'FailedAfterEvaluationWithSavedModel':
                logger.info(
                    'Model training failed, but we have a checkpoint that can be saved'
                )
                logger.info('Do you wish to save checkpoint (Y/N)?')
                answer = None
                while answer not in ['Y', 'N', 'y', 'n']:
                    answer = input()
                    if answer in ['Y', 'y']:
                        params = {'team_id': _api.team_id}
                        json_req = {'training_status': 6}
                        response = _api.send_request(
                            req_type='PUT',
                            path=f'ml_model/{new_model_id}',
                            params=params,
                            json_req=json_req
                        )
                        logger.info("Model was successfully saved")
                        pass
                    else:
                        delete_model(name)
                        logger.info('The model was not saved')
                is_training_finished = True

        time.sleep(5)
    return new_model


@model_metadata
def stop_model_training(model):
    '''This function will stop training model provided by either name or metadata, and return the ID
    :param model: The name or the metadata of the model the training of which the user needs to terminate
    :type model: str or dict
    :out model: the metadata of the now, stopped model
    :rtype out: dict
    '''
    params = {"team_id": _api.team_id}
    response = _api.send_request(
        req_type="POST",
        path=f"/ml_model/{model['id']}/stopTrainingJob",
        params=params
    )

    if response.ok:
        logger.info("Stopped model training")
    else:
        logger.info("Failed to stop model training please try again")
    return model


def plot_model_metrics(metric_json_list):
    """plots the metrics generated by neural network using plotly
       :param metric_json_list: list of <model_name>.json files
       :type  metric_json_list: list of str
    """
    def plot_df(df, plottable_cols, figure, start_index=1):
        for row, metric in enumerate(plottable_cols, start_index):
            for model_df in df:
                name = model_df['model'].iloc[0]
                x_ = model_df.loc[model_df['model'] == name, 'iteration']
                y_ = model_df.loc[model_df['model'] == name, metric]
                figure.add_trace(
                    go.Scatter(x=x_, y=y_, name=name + " " + metric),
                    row=row,
                    col=1
                )

        return figure

    def get_plottable_cols(df):
        plottable_cols = []
        for sub_df in df:
            col_names = sub_df.columns.values.tolist()
            plottable_cols += [
                col_name
                for col_name in col_names if col_name not in plottable_cols and
                col_name not in NON_PLOTABLE_KEYS
            ]
        return plottable_cols

    if not isinstance(metric_json_list, list):
        metric_json_list = [metric_json_list]

    full_c_metrics = []
    full_pe_metrics = []
    for metric_json in metric_json_list:
        data = None

        with open(metric_json) as fp:
            data = json.load(fp)
        name = metric_json.split('.')[0]
        c_metrics, pe_metrics = reformat_metrics_json(data, name)
        full_c_metrics.append(c_metrics)
        full_pe_metrics.append(pe_metrics)

    num_rows = 0
    plottable_c_cols = get_plottable_cols(full_c_metrics)
    plottable_pe_cols = get_plottable_cols(full_pe_metrics)
    num_rows = len(plottable_c_cols) + len(plottable_pe_cols)
    figure_specs = make_plotly_specs(num_rows)
    num_sublots_per_row = len(metric_json_list)
    plottable_cols = plottable_c_cols + plottable_pe_cols
    figure = make_subplots(
        rows=num_rows,
        cols=1,
        specs=figure_specs,
        subplot_titles=plottable_cols,
    )
    figure.update_layout(height=1000 * num_rows)
    models = [os.path.basename(x).split('.')[0] for x in metric_json_list]

    plot_df(full_c_metrics, plottable_c_cols, figure)
    plot_df(
        full_pe_metrics, plottable_pe_cols, figure,
        len(plottable_c_cols) + 1
    )
    figure.show()


@model_metadata
def download_model(model, output_dir):
    """Downloads the neural network and related files
    which are the <model_name>.pth/pkl. <model_name>.json, <model_name>.yaml, classes_mapper.json
    :param model: the model that needs to be downloaded
    :type  model: str
    :param output_dir: the directiory in which the files will be saved
    :type output_dir: str
    :out model: the metadata of the model that was deleted
    :rtype model: dict
    """
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir, exist_ok=True)

    weights_path = model["path"]
    weights_name = os.path.basename(weights_path)
    metrics_name = weights_name.split('.')[0] + '.json'

    config_path = model["config_path"]
    _path = config_path.split("/")
    _path[-1] = "classes_mapper.json"
    mapper_path = "/".join(_path)
    _path[-1] = metrics_name
    metrics_path = "/".join(_path)

    params = {
        "team_id": _api.team_id,
    }

    response = _api.send_request(
        req_type="GET",
        path=f"/ml_model/getMyModelDownloadToken/{model['id']}",
        params=params
    )
    if response.ok:
        response = process_api_response(response.json())
    else:
        raise SABaseException(0, "Could not get model info ")

    tokens = response["tokens"]
    s3_session = boto3.Session(
        aws_access_key_id=tokens["accessKeyId"],
        aws_secret_access_key=tokens["secretAccessKey"],
        aws_session_token=tokens["sessionToken"]
    )
    s3_resource = s3_session.resource('s3')

    bucket = s3_resource.Bucket(tokens["bucket"])

    bucket.download_file(config_path, os.path.join(output_dir, 'config.yaml'))
    bucket.download_file(weights_path, os.path.join(output_dir, weights_name))
    try:
        bucket.download_file(
            metrics_path, os.path.join(output_dir, metrics_path)
        )
        bucket.download_file(mapper_path, os.path.join(output_dir, mapper_path))

    except Exception as e:
        logger.info(
            "the specified model does not contain a classes_mapper and/or a metrics file"
        )

    logger.info("Downloaded model related files")
    return model


@model_metadata
def delete_model(model):
    '''This function deletes the provided model
       :param model: the model to be deleted
       :out model: the metadata of the model that was deleted
       :rtype out: dict
    '''
    params = {"team_id": _api.team_id}
    response = _api.send_request(
        req_type="DELETE", path=f'/ml_model/{model["id"]}', params=params
    )
    if response.ok:
        logger.info("Model successfully deleted")
        raise SABaseException(0, 'Failed to delete model')
    else:
        logger.info("Failed to delete model, please try again")
    return model



DEFAULT_HYPERPARAMETERS = {
    "instance_type": "1 x T4 16 GB",
    "num_epochs": 12,
    "dataset_split_ratio": 80,
    "base_lr": 0.02,
    "gamma": 0.5,
    "images_per_batch": 4,
    "batch_per_image": 512,
    "steps": [5],
    "evaluation_period": 12,
    "runtime_seconds": 600,
    "estimated_remaining_time": 600,
    "template_id": None
}

NON_PLOTABLE_KEYS = ['eta_seconds', 'iteration', 'data_time', 'time', 'model']
PLOTTABLE_METRICS = ['total_loss']
DROP_KEYS = ['eta_seconds', 'data_time', 'time']



'''
Sagemaker to SA conversion method
'''
import logging
import threading
import json
from pathlib import Path
import cv2
import numpy as np

from ..sa_json_helper import (_create_pixel_instance, _create_sa_json)

from ....common import (
    hex_to_rgb, blue_color_generator, write_to_json, tqdm_converter
)

logger = logging.getLogger("superannotate-python-sdk")


def sagemaker_instance_segmentation_to_sa_pixel(data_path, output_dir):
    img_mapping = {}
    try:
        img_map_file = open(data_path / 'output.manifest')
    except Exception as e:
        raise Exception("'output.manifest' file doesn't exist")

    for line in img_map_file:
        dd = json.loads(line)
        img_mapping[Path(dd['attribute-name-ref']).name] = Path(
            dd['source-ref']
        ).name

    json_list = list(data_path.glob('*.json'))
    classes_ids = {}
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_list), images_converted, images_not_converted, finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()
    for json_file in json_list:
        data_json = json.load(open(json_file))
        for annotataion in data_json:
            if 'consolidatedAnnotation' not in annotataion.keys():
                logger.warning('Wrong json files')
                raise Exception

            mask_name = Path(
                annotataion['consolidatedAnnotation']['content']
                ['attribute-name-ref']
            ).name

            classes_dict = annotataion['consolidatedAnnotation']['content'][
                'attribute-name-ref-metadata']['internal-color-map']

            try:
                img = cv2.imread(str(data_path / mask_name.replace(':', '_')))
                H, W, _ = img.shape
            except Exception as e:
                logger.warning(
                    "Can't open %s mask", mask_name.replace(':', '_')
                )
                images_not_converted.append(mask_name.replace(':', '_'))
                continue

            class_contours = {}
            num_of_contours = 0
            for key, _ in classes_dict.items():
                if classes_dict[key]['class-name'] == 'BACKGROUND':
                    continue

                if key not in classes_ids.keys():
                    classes_ids[key] = classes_dict[key]['class-name']

                bitmask = np.zeros((H, W), dtype=np.int8)
                bitmask[np.all(
                    img == list(hex_to_rgb(classes_dict[key]['hex-color'])
                               )[::-1],
                    axis=2
                )] = 255

                bitmask = bitmask.astype(np.uint8)
                contours, _ = cv2.findContours(
                    bitmask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
                )
                class_contours[classes_dict[key]['class-name']] = contours
                num_of_contours += len(contours)

            blue_colors = blue_color_generator(num_of_contours)
            idx = 0
            file_name = '%s___pixel.json' % (img_mapping[mask_name])
            sa_metadata = {
                'name': img_mapping[mask_name],
                'width': W,
                'height': H
            }
            sa_instances = []
            sa_mask = np.zeros((H, W, 4))
            for name, contours in class_contours.items():
                parts = []
                for contour in contours:
                    bitmask = np.zeros((H, W))
                    contour = contour.flatten().tolist()
                    pts = np.array(
                        [
                            contour[2 * i:2 * (i + 1)]
                            for i in range(len(contour) // 2)
                        ],
                        dtype=np.int32
                    )
                    cv2.fillPoly(bitmask, [pts], 1)
                    sa_mask[bitmask == 1
                           ] = list(hex_to_rgb(blue_colors[idx]))[::-1] + [255]
                    parts.append({'color': blue_colors[idx]})
                    idx += 1
                    sa_obj = _create_pixel_instance(parts, [], name)
                    sa_instances.append(sa_obj)
            images_converted.append(img_mapping[mask_name])
            sa_json = _create_sa_json(sa_instances, sa_metadata)
            write_to_json(output_dir / file_name, sa_json)
            cv2.imwrite(
                str(output_dir / (img_mapping[mask_name] + '___save.png')),
                sa_mask
            )
    finish_event.set()
    tqdm_thread.join()
    return classes_ids



from pathlib import Path
import superannotate as sa
from ..common import upload_project


def test_vgg_convert_object(tmpdir):
    project_name = "vgg_test_object"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "VGG" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VGG", "vgg_test", "Vector", "object_detection"
    )

    description = 'vgg object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_vgg_convert_instance(tmpdir):
    project_name = "vgg_test_instance"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "VGG" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VGG", "vgg_test", "Vector", "instance_segmentation"
    )

    description = 'vgg instance segmentation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_vgg_convert_vector(tmpdir):
    project_name = "vgg_test_vector"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "VGG" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VGG", "vgg_test", "Vector", "vector_annotation"
    )

    description = 'vgg vector annotation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)



from pathlib import Path

import superannotate as sa
import json


def test_coco_split(tmpdir):
    image_dir = Path(
        'tests'
    ) / 'converter_test' / 'COCO' / 'input' / 'toSuperAnnotate' / 'instance_segmentation'
    coco_json = image_dir / 'instances_test.json'
    out_dir = Path(tmpdir) / "coco_split"

    sa.coco_split_dataset(
        coco_json, image_dir, out_dir, ['split1', 'split2', 'split3'],
        [50, 30, 20]
    )

    main_json = json.load(open(coco_json))
    split1_json = json.load(open(out_dir / 'split1.json'))
    split2_json = json.load(open(out_dir / 'split2.json'))
    split3_json = json.load(open(out_dir / 'split3.json'))

    assert len(main_json['images']) == (
        len(split1_json['images']) + len(split2_json['images']) +
        len(split3_json['images'])
    )
    assert len(main_json['annotations']) == (
        len(split1_json['annotations']) + len(split2_json['annotations']) +
        len(split3_json['annotations'])
    )


'''
Supervisely to SA conversion method
'''
import logging
import threading
import json
from pathlib import Path
import cv2
import numpy as np

from .supervisely_helper import _base64_to_polygon, _create_attribute_list

from ..sa_json_helper import _create_pixel_instance, _create_sa_json

from ....common import (
    hex_to_rgb, blue_color_generator, write_to_json, tqdm_converter
)

logger = logging.getLogger("superannotate-python-sdk")


def supervisely_instance_segmentation_to_sa_pixel(
    json_files, class_id_map, output_dir
):
    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(
            len(json_files), images_converted, images_not_converted,
            finish_event
        ),
        daemon=True
    )
    logger.info('Converting to SuperAnnotate JSON format')
    tqdm_thread.start()

    for json_file in json_files:
        file_name = '%s___pixel.json' % Path(json_file).stem

        json_data = json.load(open(json_file))
        sa_instances = []

        H, W = json_data['size']['height'], json_data['size']['width']
        mask = np.zeros((H, W, 4))
        sa_metadata = {'name': Path(json_file).stem, 'width': W, 'height': H}

        hex_colors = blue_color_generator(10 * len(json_data['objects']))
        index = 0
        for obj in json_data['objects']:
            if 'classTitle' in obj and obj['classTitle'] in class_id_map.keys():
                attributes = []
                if 'tags' in obj.keys():
                    attributes = _create_attribute_list(
                        obj['tags'], obj['classTitle'], class_id_map
                    )
                    parts = []
                    if obj['geometryType'] == 'bitmap':
                        segments = _base64_to_polygon(obj['bitmap']['data'])
                        for segment in segments:
                            ppoints = [
                                x + obj['bitmap']['origin'][0] if i %
                                2 == 0 else x + obj['bitmap']['origin'][1]
                                for i, x in enumerate(segment)
                            ]
                            bitmask = np.zeros((H, W)).astype(np.uint8)
                            pts = np.array(
                                [
                                    ppoints[2 * i:2 * (i + 1)]
                                    for i in range(len(ppoints) // 2)
                                ],
                                dtype=np.int32
                            )
                            cv2.fillPoly(bitmask, [pts], 1)
                            color = hex_to_rgb(hex_colors[index])
                            mask[bitmask == 1] = list(color[::-1]) + [255]
                            parts.append({'color': hex_colors[index]})
                            index += 1
                        cv2.imwrite(
                            str(
                                output_dir / file_name.
                                replace('___pixel.json', '___save.png')
                            ), mask
                        )
                        sa_obj = _create_pixel_instance(
                            parts, attributes, obj['classTitle']
                        )
                        sa_instances.append(sa_obj)

        images_converted.append(Path(json_file).stem)
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)
    finish_event.set()
    tqdm_thread.join()



import logging
import requests

logger = logging.getLogger("superannotate-python-sdk")


def image_downloader(url, file_name):
    r = requests.get(url, stream=True)
    if r.status_code == 200:
        with open(file_name, 'wb') as f:
            f.write(r.content)
        return True
    return False


def _create_classes_id_map(json_data):
    classes = {}
    for d in json_data:
        if 'objects' not in d['Label'].keys():
            continue

        instances = d["Label"]["objects"]
        for instance in instances:
            class_name = instance["value"]
            if class_name not in classes.keys():
                color = instance["color"]
                classes[class_name] = {"color": color, 'attribute_groups': {}}

            if 'classifications' in instance.keys():
                classifications = instance['classifications']
                for classification in classifications:
                    if classification['value'] not in classes[class_name][
                        'attribute_groups']:
                        if 'answer' in classification.keys():
                            if isinstance(classification['answer'], str):
                                continue

                            classes[class_name]['attribute_groups'][
                                classification['value']] = {
                                    'is_multiselect': 0,
                                    'attributes': []
                                }
                            if isinstance(classification['answer'], list):
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(
                                    classification['answer'][0]['value']
                                )
                            elif isinstance(classification['answer'], dict):
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(
                                    classification['answer']['value']
                                )

                        elif 'answers' in classification.keys():
                            classes[class_name]['attribute_groups'][
                                classification['value']] = {
                                    'is_multiselect': 1,
                                    'attributes': []
                                }
                            for attr in classification['answers']:
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(attr['value'])

                    else:
                        if 'answer' in classification.keys():
                            if isinstance(classification['answer'], dict):
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(
                                    classification['answer']['value']
                                )
                            elif isinstance(classification['answer'], list):
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(
                                    classification['answer'][0]['value']
                                )
                        elif 'answers' in classification.keys():
                            for attr in classification['answers']:
                                classes[class_name]['attribute_groups'][
                                    classification['value']
                                ]['attributes'].append(attr['value'])
    return classes


def _create_attributes_list(lb_attributes):
    attributes = []
    for attribute in lb_attributes:
        group_name = attribute['value']
        if 'answer' in attribute.keys(
        ) and isinstance(attribute['answer'], dict):
            attribute_name = attribute['answer']['value']
            attr_dict = {'name': attribute_name, 'groupName': group_name}
            attributes.append(attr_dict)
        elif 'answers' in attribute.keys():
            for attr in attribute['answers']:
                attribute_name = attr['value']
                attr_dict = {'name': attribute_name, 'groupName': group_name}
                attributes.append(attr_dict)

    return attributes



import numpy as np

from ..baseStrategy import baseStrategy

from ....common import write_to_json


class VGGStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        json_data = self.export_root / (self.dataset_name + '.json')
        classes = self.conversion_algorithm(
            json_data, self.task, self.output_dir
        )
        sa_classes = self._create_classes(classes)
        (self.output_dir / 'classes').mkdir(exist_ok=True)
        write_to_json(self.output_dir / 'classes' / 'classes.json', sa_classes)

    def _create_classes(self, class_id_map):
        sa_classes = []
        for key in class_id_map.keys():
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            dd = {'name': key, 'color': hexcolor, 'attribute_groups': []}
            for attributes, value in class_id_map[key]['attribute_groups'
                                                      ].items():
                attr_group = {
                    'name': attributes,
                    'is_multiselect': value['is_multiselect'],
                    'attributes': []
                }
                for attribute in value['attributes']:
                    attr = {'name': attribute, 'groupName': attributes}
                    attr_group['attributes'].append(attr.copy())
                dd['attribute_groups'].append(attr_group.copy())
            sa_classes.append(dd)
        return sa_classes



import superannotate as sa
from pathlib import Path
import json


def compare_jsons(json_gen, input_dir):
    for path in json_gen:
        final_json = json.load(open(str(path)))
        input_path = input_dir.joinpath(path.name)
        init_json = json.load(open(str(input_path)))

        for init, final in zip(init_json['instances'], final_json['instances']):
            for key in init.keys():
                if key == 'parts':
                    continue
                init_value = init[key]
                final_value = final[key]
                assert init_value == final_value


def test_pixel_vector_pixel(tmpdir):
    input_dir = Path()
    input_dir = input_dir.joinpath(
        'tests', 'converter_test', 'COCO', 'input', 'fromSuperAnnotate',
        'cats_dogs_pixel_instance_segm'
    )
    temp_dir = Path(tmpdir) / 'output'
    final_dir = Path(tmpdir) / 'output2'

    sa.convert_project_type(input_dir, temp_dir)
    sa.convert_project_type(temp_dir, final_dir)

    gen = final_dir.glob('*.json')
    compare_jsons(gen, input_dir)


def test_vector_pixel_vector(tmpdir):
    input_dir = Path(
        'tests'
    ) / 'converter_test' / 'COCO' / 'input' / 'fromSuperAnnotate' / 'cats_dogs_vector_instance_segm'
    temp_dir = Path(tmpdir) / 'output'
    final_dir = Path(tmpdir) / 'output2'

    sa.convert_project_type(input_dir, temp_dir)
    sa.convert_project_type(temp_dir, final_dir)

    gen = input_dir.glob('*.json')
    compare_jsons(gen, input_dir)


from pathlib import Path
import superannotate as sa
from ..common import upload_project


def test_vott_convert_object(tmpdir):
    project_name = "vott_object"
    input_dir = Path(
        "tests"
    ) / "converter_test" / "VoTT" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VoTT", "", "Vector", "object_detection"
    )

    description = 'vott object detection'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_vott_convert_instance(tmpdir):
    project_name = "vott_vector_instance"
    input_dir = Path(
        "tests"
    ) / "converter_test" / "VoTT" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VoTT", "", "Vector", "instance_segmentation"
    )

    description = 'vott instance segmentation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)


def test_vott_convert_vector(tmpdir):
    project_name = "vott_vector_annotation"

    input_dir = Path(
        "tests"
    ) / "converter_test" / "VoTT" / "input" / "toSuperAnnotate"
    out_dir = Path(tmpdir) / project_name
    sa.import_annotation(
        input_dir, out_dir, "VoTT", "", "Vector", "vector_annotation"
    )

    description = 'vott vector annotation'
    ptype = 'Vector'
    upload_project(out_dir, project_name, description, ptype)



import sys

from setuptools import find_packages, setup

with open('requirements.txt') as f:
    requirements = f.read()
requirements = requirements.splitlines()

if sys.platform == 'linux':
    with open('requirements_extra.txt') as f:
        requirements_extra = f.read()

    requirements_extra = requirements_extra.splitlines()
    requirements += requirements_extra

with open('README.md') as f:
    readme = f.read()
readme = "\n".join(readme.split('\n')[2:])

packages = find_packages()

with open('superannotate/version.py') as f:
    Version = f.read()

Version = Version.rstrip()
Version = Version[15:-1]

setup(
    name='superannotate',
    version=Version,
    description='Python SDK to SuperAnnotate platform',
    license='MIT',
    author='SuperAnnotate AI',
    author_email='hovnatan@superannotate.com',
    url='https://github.com/superannotateai/superannotate-python-sdk',
    long_description=readme,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    setup_requires=['wheel'],
    packages=find_packages(exclude=('tests', )),
    entry_points={
        'console_scripts': ['superannotatecli = superannotate.__main__:main']
    },
    python_requires='>=3.6'
)



from pathlib import Path
import numpy as np

from ..baseStrategy import baseStrategy

from ....common import write_to_json


class GoogleCloudStrategy(baseStrategy):
    def __init__(self, args):
        super().__init__(args)

    def to_sa_format(self):
        path = Path(self.export_root).joinpath(self.dataset_name + '.csv')
        classes = self.conversion_algorithm(path, self.output_dir)
        sa_classes = self._create_classes(classes)
        (self.output_dir / 'classes').mkdir(exist_ok=True)
        write_to_json(self.output_dir / 'classes' / 'classes.json', sa_classes)

    def _create_classes(self, classes):
        classes_loader = []
        for class_ in set(classes):
            color = np.random.choice(range(256), size=3)
            hexcolor = "#%02x%02x%02x" % tuple(color)
            sa_classes = {
                'name': class_,
                'color': hexcolor,
                'attribute_groups': []
            }
            classes_loader.append(sa_classes)
        return classes_loader



import copy
import logging

from .annotation_classes import search_annotation_classes
from .project_api import (
    get_project_metadata_bare, get_project_metadata_with_users
)
from .projects import get_project_settings, get_project_workflow

logger = logging.getLogger("superannotate-python-sdk")


def get_project_metadata(
    project,
    include_annotation_classes=False,
    include_settings=False,
    include_workflow=False,
    include_contributors=False,
    include_complete_image_count=False
):
    """Returns project metadata

    :param project: project name or project metadata from previous calls
    :type project: str or dict
    :param include_annotation_classes: enables project annotation classes output under
                                       the key "annotation_classes"
    :type include_annotation_classes: bool
    :param include_settings: enables project settings output under
                             the key "settings"
    :type include_settings: bool
    :param include_workflow: enables project workflow output under
                             the key "workflow"
    :type include_workflow: bool
    :param include_contributors: enables project contributors output under
                             the key "contributors"
    :type include_contributors: bool

    :return: metadata of project
    :rtype: dict
    """
    if not isinstance(project, dict):
        project = get_project_metadata_bare(
            project, include_complete_image_count
        )
    result = copy.deepcopy(project)
    if include_annotation_classes:
        result["annotation_classes"] = search_annotation_classes(project)
    if include_contributors:
        result["contributors"] = get_project_metadata_with_users(project
                                                                )["users"]
    if include_settings:
        result["settings"] = get_project_settings(project)
    if include_workflow:
        result["workflow"] = get_project_workflow(project)
    return result



import json
import logging
from pathlib import Path

import packaging.version
import requests

from .analytics.class_analytics import (
    attribute_distribution, class_distribution
)
from .analytics.common import aggregate_annotations_as_df, df_to_annotations
from .annotation_helpers import (
    add_annotation_bbox_to_json, add_annotation_comment_to_json,
    add_annotation_cuboid_to_json, add_annotation_ellipse_to_json,
    add_annotation_point_to_json, add_annotation_polygon_to_json,
    add_annotation_polyline_to_json, add_annotation_template_to_json
)
from .api import API as _API
from .common import image_path_to_annotation_paths

try:
    from .consensus_benchmark.benchmark import benchmark
    from .consensus_benchmark.consensus import consensus
except:
    _WARNING_SHAPELY = "To use superannotate.benchmark or superannotate.consensus functions please install shapely package in Anaconda enviornment with # conda install shapely"

    def benchmark(*args, **kwargs):
        raise RuntimeError(_WARNING_SHAPELY)

    def consensus(*args, **kwargs):
        raise RuntimeError(_WARNING_SHAPELY)


from .dataframe_filtering import (
    filter_annotation_instances, filter_images_by_comments,
    filter_images_by_tags
)
from .db.annotation_classes import (
    create_annotation_class, create_annotation_classes_from_classes_json,
    delete_annotation_class, download_annotation_classes_json,
    get_annotation_class_metadata, search_annotation_classes
)
from .db.clone_project import clone_project
from .db.exports import (
    download_export, get_export_metadata, get_exports, prepare_export
)
from .db.images import (
    add_annotation_bbox_to_image, add_annotation_comment_to_image,
    add_annotation_cuboid_to_image, add_annotation_ellipse_to_image,
    add_annotation_point_to_image, add_annotation_polygon_to_image,
    add_annotation_polyline_to_image, add_annotation_template_to_image,
    create_fuse_image, delete_image, download_image, download_image_annotations,
    download_image_preannotations, get_image_annotations, get_image_bytes,
    get_image_metadata, get_image_preannotations, search_images,
    set_image_annotation_status, upload_image_annotations
)
from .db.project_images import (
    assign_images, copy_image, move_image, pin_image, upload_image_to_project
)
from .db.project_metadata import get_project_metadata
from .db.projects import (
    create_project, create_project_from_metadata, delete_project,
    get_project_default_image_quality_in_editor, get_project_image_count,
    get_project_settings, get_project_workflow, rename_project,
    set_project_default_image_quality_in_editor, set_project_settings,
    set_project_workflow, share_project, unshare_project,
    upload_annotations_from_folder_to_project,
    upload_images_from_azure_blob_to_project,
    upload_images_from_folder_to_project,
    upload_images_from_google_cloud_to_project,
    upload_images_from_public_urls_to_project,
    upload_images_from_s3_bucket_to_project, upload_images_to_project,
    upload_preannotations_from_folder_to_project, upload_video_to_project,
    upload_videos_from_folder_to_project
)
from .db.search_projects import search_projects
from .db.teams import (
    delete_contributor_to_team_invitation, get_team_metadata,
    invite_contributor_to_team
)
from .db.users import search_team_contributors
from .dicom_converter import dicom_to_rgb_sequence
from .exceptions import (
    SABaseException, SAExistingAnnotationClassNameException,
    SAExistingProjectNameException, SANonExistingAnnotationClassNameException,
    SANonExistingProjectNameException
)
from .input_converters.conversion import (
    coco_split_dataset, convert_project_type, export_annotation,
    import_annotation, convert_json_version
)
from .ml.ml_funcs import (
    delete_model, download_model, plot_model_metrics, run_prediction,
    run_segmentation, run_training, stop_model_training
)
from .ml.ml_models import search_models

from .old_to_new_format_convertor import update_json_format
from .version import __version__

formatter = logging.Formatter(fmt='SA-PYTHON-SDK - %(levelname)s - %(message)s')
#formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

handler = logging.StreamHandler()
handler.setFormatter(formatter)

logger = logging.getLogger("superannotate-python-sdk")
logger.setLevel(logging.INFO)
logger.addHandler(handler)


def _check_version():
    local_version = packaging.version.parse(__version__)
    if local_version.is_prerelease:
        logger.warning(
            "Development version %s of SuperAnnotate SDK is being used",
            __version__
        )
    req = requests.get('https://pypi.python.org/pypi/superannotate/json')
    if not req.ok:
        return

    # find max version on PyPI
    releases = req.json().get('releases', [])
    pip_version = packaging.version.parse('0')
    for release in releases:
        ver = packaging.version.parse(release)
        if not ver.is_prerelease or local_version.is_prerelease:
            pip_version = max(pip_version, ver)

    if pip_version.major > local_version.major:
        logger.warning(
            "There is a major upgrade of SuperAnnotate Python SDK available on PyPI. We recommend upgrading. Run 'pip install --upgrade superannotate' to upgrade from your version %s to %s.",
            local_version, pip_version
        )
    elif pip_version > local_version:
        logger.info(
            "There is a newer version of SuperAnnotate Python SDK available on PyPI. Run 'pip install --upgrade superannotate' to upgrade from your version %s to %s.",
            local_version, pip_version
        )


_api = _API.get_instance()


def init(path_to_config_json):
    """Initializes and authenticates to SuperAnnotate platform using the config file.
    If not initialized then $HOME/.superannotate/config.json
    will be used.

    :param path_to_config_json: Location to config JSON file
    :type path_to_config_json: str or Path
    """
    _api.init(path_to_config_json)


_check_version()
init(None)



from pathlib import Path

import superannotate as sa


# COCO to SA
# test panoptic-segmentation
def test_panoptic_segmentation_coco2sa(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "panoptic_segmentation"
    out_path = Path(tmpdir) / "toSuperAnnotate" / "panoptic_test"
    sa.import_annotation(
        input_dir, out_path, "COCO", "panoptic_test", "Pixel",
        "panoptic_segmentation"
    )


# test keypoint-detection
def test_keypoint_detection_coco2sa(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "keypoint_detection"
    out_path = Path(tmpdir) / "toSuperAnnotate" / "keypoint_test"
    sa.import_annotation(
        input_dir, out_path, "COCO", "person_keypoints_test", "Vector",
        "keypoint_detection"
    )


# test instance segmentation
def test_instance_segmentation_coco2sa(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "toSuperAnnotate" / "instance_segmentation"
    out_path = Path(tmpdir) / "toSuperAnnotate" / "instances_test"
    sa.import_annotation(
        input_dir, out_path, "COCO", "instances_test", "Vector",
        "instance_segmentation"
    )


# SA to COCO
# test panoptic segmentation
def test_panoptic_segmentation_sa2coco(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "fromSuperAnnotate" / "cats_dogs_panoptic_segm"
    out_path = Path(tmpdir) / "fromSuperAnnotate" / "panoptic_test"
    sa.export_annotation(
        input_dir, out_path, "COCO", "panoptic_test", "Pixel",
        "panoptic_segmentation"
    )


def test_keypoint_detection_sa2coco(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "fromSuperAnnotate" / "cats_dogs_vector_keypoint_det"
    out_path = Path(tmpdir) / "fromSuperAnnotate" / "keypoint_test_vector"
    sa.export_annotation(
        input_dir, out_path, "COCO", "keypoint_test_vector", "Vector",
        "keypoint_detection"
    )


def test_instance_segmentation_sa2coco_pixel(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "fromSuperAnnotate" / "cats_dogs_pixel_instance_segm"
    out_path = Path(tmpdir) / "fromSuperAnnotate" / "instance_test_pixel"
    sa.export_annotation(
        input_dir, out_path, "COCO", "instance_test_pixel", "Pixel",
        "instance_segmentation"
    )


def test_instance_segmentation_sa2coco_vector(tmpdir):
    input_dir = Path(
        "tests"
    ) / "converter_test" / "COCO" / "input" / "fromSuperAnnotate" / "cats_dogs_vector_instance_segm"
    out_path = Path(tmpdir) / "fromSuperAnnotate" / "instance_test_vector"
    sa.export_annotation(
        input_dir, out_path, "COCO", "instance_test_vector", "Vector",
        "instance_segmentation"
    )
